/* --------------------------------------------------------------------------
 *
 * License
 *
 * The contents of this file are subject to the Jabber Open Source License
 * Version 1.0 (the "License").  You may not copy or use this file, in either
 * source code or executable form, except in compliance with the License.  You
 * may obtain a copy of the License at http://www.jabber.com/license/ or at
 * http://www.opensource.org/.  
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied.  See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * Copyrights
 * 
 * Portions created by or assigned to Jabber.com, Inc. are 
 * Copyright (c) 1999-2000 Jabber.com, Inc.  All Rights Reserved.  Contact
 * information for Jabber.com, Inc. is available at http://www.jabber.com/.
 *
 * Portions Copyright (c) 1998-1999 Jeremie Miller.
 * 
 * Acknowledgements
 * 
 * Special thanks to the Jabber Open Source Contributors for their
 * suggestions and support of Jabber.
 * 
 * --------------------------------------------------------------------------*/
 
#include "http.h"


void _http_purge(xht h, const char *key, void *data, void *arg)
{
    hrq rq = (hrq)data;
    int now = time(NULL);

    if(rq->keepalive != NULL)
        mio_write(rq->m,NULL,rq->keepalive,-1);

    if(rq->phase >= PHASE_DONE) return;

    if((now - rq->stamp) < rq->h->timeout) return;

    log_debug(ZONE,"timing out http request from %s",mio_ip(rq->m));
    mio_close(rq->m);
}

/* walk the table looking for stale requests to expire */
result http_purge(void *arg)
{
    hpi hi = (hpi)arg;
    xhash_walk(hi->readers,_http_purge,NULL);
    return r_DONE;
}

/* the callback to handle incoming packets */
result http_packets(instance i, dpacket p, void *arg)
{
    hpi h = (hpi)arg;
    xmlnode x;

    if(p->type != p_ROUTE) return r_ERR; /* we only handle routed stuff! */

    /* outgoing requests need to have their dns resolved */
    if(j_strcmp(xmlnode_get_attrib(p->x,"type"),"request") == 0 && (x = xmlnode_get_firstchild(p->x)) != NULL)
    {
        if(h->dnsrv == NULL)
        {
            deliver_fail(p,"No DNS Resolver Configured");
            return r_DONE;
        }
        /* preserve the from atrib, and wrap for dnsrv */
        xmlnode_put_attrib(x,"from",xmlnode_get_attrib(p->x,"from"));
        x = xmlnode_wrap(x,"route");
        xmlnode_put_attrib(x,"to",h->dnsrv);
        log_debug(ZONE,"routing to dns: %s",xmlnode2str(x));
        deliver(dpacket_new(x),NULL);
        return r_DONE;
    }

    /* things with an ip are new outgoing requests */
    if(xmlnode_get_attrib(p->x,"ip") != NULL)
        http_out(h, p);
    else
        http_in(h, p);

    return r_DONE;
}


void http(instance i, xmlnode x)
{
    char *host;
    xmlnode config, map, cur;
    xdbcache xc;
    hpi hi;
    int port;
    hl l;
    mio m;

    log_debug(ZONE,"http loading");

    xc = xdb_cache(i);
    config = xdb_get(xc, jid_new(xmlnode_pool(x),"config@-internal"),"jabber:config:http");

    /* set up internal instance and hash tables */
    hi = pmalloco(i->p,sizeof(_hpi));
    hi->readers = xhash_new(j_atoi(xmlnode_get_tag_data(config,"maxrequests"),67));
    hi->i = i;
    hi->timeout = j_atoi(xmlnode_get_tag_data(config,"timeout"),30);
    hi->dnsrv = pstrdup(i->p,xmlnode_get_tag_data(config,"dnsrv"));
    hi->proxy = pstrdup(i->p,xmlnode_get_tag_data(config,"proxy"));
    hi->proxyp = j_atoi(xmlnode_get_attrib(xmlnode_get_tag(config,"proxy"),"port"),80);

    /* parse the <listen> elements */
    for(cur = xmlnode_get_tag(config,"listen"); cur != NULL; xmlnode_hide(cur), cur = xmlnode_get_tag(config,"listen"))
    {
        if((port = j_atoi(xmlnode_get_attrib(cur,"port"),0)) <= 0) continue;

        l = pmalloco(i->p,sizeof(_hl));
        l->h = hi;
        l->map = xhash_new(j_atoi(xmlnode_get_tag_data(config,"maxhosts"),11));
        /* parse the <map> elements */
        for(map = xmlnode_get_tag(cur,"map"); map != NULL; xmlnode_hide(map), map = xmlnode_get_tag(cur,"map"))
        {
            if((host = xmlnode_get_data(map)) != NULL)
                xhash_put(l->map, pstrdup(i->p, host), pstrdup(i->p, xmlnode_get_attrib(map,"to")));
            else
                l->def = pstrdup(i->p, xmlnode_get_attrib(map,"to"));
        }
        if((m = mio_listen(port, xmlnode_get_attrib(cur,"ip"), http_in_connect, (void*)l, MIO_LISTEN_RAW)) == NULL)
            log_error(i->id,"failed to listen on %s:%d",xmlnode_get_attrib(cur,"ip"),port);
        else
            mio_rate(m, 2, 100);
    }

    register_phandler(i, o_DELIVER, http_packets, (void *)hi);
    if(hi->timeout > 0) /* 0 means never timeout */
        register_beat(hi->timeout, http_purge, (void *)hi);

    xmlnode_free(config);
}

