/* --------------------------------------------------------------------------
 *
 * License
 *
 * The contents of this file are subject to the Jabber Open Source License
 * Version 1.0 (the "License").  You may not copy or use this file, in either
 * source code or executable form, except in compliance with the License.  You
 * may obtain a copy of the License at http://www.jabber.com/license/ or at
 * http://www.opensource.org/.  
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied.  See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * Copyrights
 * 
 * Portions created by or assigned to Jabber.com, Inc. are 
 * Copyright (c) 1999-2000 Jabber.com, Inc.  All Rights Reserved.  Contact
 * information for Jabber.com, Inc. is available at http://www.jabber.com/.
 *
 * Portions Copyright (c) 1998-1999 Jeremie Miller.
 * 
 * Acknowledgements
 * 
 * Special thanks to the Jabber Open Source Contributors for their
 * suggestions and support of Jabber.
 * 
 * --------------------------------------------------------------------------*/
 
#include "jabberd.h"

#undef VERSION
#define VERSION "0.4"

/*******

packet parser looks for to="hashid" and calls http_in(dpacket)
    else look for ip="1.2.3.4" and calls http_out(dpacket)

void http_read(mio m, callback)
    resets read to internally
    internal timeout
    callback with error flag, then xmlnode is <error>FooBar</error>

http_in
    find hashid, else bounce w/ error (unless error already)
    if error, disconnect
    write partial or full response

http_out
    use ip and connect to it
    failing, bounce
    write request
    http_read() result
    modify original packet and reply

*******/

/* http parser instance */
typedef struct hpi_struct
{
    instance i;
    xht readers; /* connections being handled/parsed */
    int timeout; /* how long to wait for an http request to be completly sent or received */
    char *dnsrv; /* our dns resolver */
    char *proxy; /* outgoing proxy */
    int proxyp;  /* port for proxy */
} *hpi, _hpi;

/* http listener */
typedef struct hl_struct
{
    hpi h;
    xht map; /* virtual hostnames mapping to a component name */
    char *def; /* default component */
} *hl, _hl;

/*** http parser "reader"
 *
 * the reader takes a socket and attempts to parse the http protocol off of it
 * it'll parse either an incoming request or the result after an outgoing one
 * it takes a function callback, which is called either after an error has occurred or the parsing is done
 * this is reflected in the phase, as DONE or ERROR
 */
#define PHASE_START 0
#define PHASE_HEAD 1
#define PHASE_BODY 2
#define PHASE_DONE 3
#define PHASE_ERROR 4

typedef struct hrq_struct *hrq, _hrq;

typedef void (*httpreader)(hrq rq);

/* http reader parser */
struct hrq_struct
{
    int phase;
    char id[16];
    mio m;
    hpi h;
    int clen, stamp;
    char *buffer, *keepalive;
    xmlnode x, head, body;
    pool p;
    httpreader f;
    hl l;
};

hrq http_read(hpi h, mio m, httpreader f);


/* stubs for calls between files */
void http_in(hpi h, dpacket p);
void http_in_connect(mio m, int flag, void *arg, char *buffer, int bufsz);
void http_out(hpi h, dpacket p);
