/* --------------------------------------------------------------------------
 *
 * License
 *
 * The contents of this file are subject to the Jabber Open Source License
 * Version 1.0 (the "License").  You may not copy or use this file, in either
 * source code or executable form, except in compliance with the License.  You
 * may obtain a copy of the License at http://www.jabber.com/license/ or at
 * http://www.opensource.org/.  
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied.  See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * Copyrights
 * 
 * Portions created by or assigned to Jabber.com, Inc. are 
 * Copyright (c) 1999-2000 Jabber.com, Inc.  All Rights Reserved.  Contact
 * information for Jabber.com, Inc. is available at http://www.jabber.com/.
 *
 * Portions Copyright (c) 1998-1999 Jeremie Miller.
 * 
 * Acknowledgements
 * 
 * Special thanks to the Jabber Open Source Contributors for their
 * suggestions and support of Jabber.
 * 
 * --------------------------------------------------------------------------*/
 
#include "http.h"


int _http_dehex(char c)
{
    if(c >= '0' && c <= '9')
        return c - '0';
    if(c >= 'a' && c <= 'f')
        return c - 'a' + 10;
    if(c >= 'A' && c <= 'F')
        return c - 'A' + 10;
    return 0;
}

/* decode the uri hex junk */
void http_strdecode(char *from)
{
    char *to;

    to = from;
    for(;*from != '\0'; ++to, ++from)
    {
        if(from[0] == '%' && strchr("0123456789abcdefABCDEF", from[1]) != NULL && strchr("0123456789abcdefABCDEF", from[2]) != NULL)
        {
            *to = _http_dehex(from[1]) * 16 + _http_dehex(from[2]);
            from += 2;
        }else{
            *to = *from;
        }
    }
    *to = '\0';
}

void http_qs2xml(xmlnode x, char *qs)
{
    char *key, *val, *nul;

    if(qs == NULL) return;

    val = nul = qs + strlen(qs);
    key = qs;
    for(;qs[0] != '\0'; ++qs)
    {
        if(qs[0] == '+')
            qs[0] = ' ';
        if(qs[0] == '=')
        {
            qs[0] = '\0';
            val = qs + 1;
        }
        if(qs[0] == '&' || qs[1] == '\0')
        {
            if(qs[0] == '&')
                qs[0] = '\0';
            http_strdecode(key);
            http_strdecode(val);
            xmlnode_put_attrib(x,key,val);
            key = qs + 1;
            val = nul;
        }
    }
}


/* xml-ize common http stuff */
void http_massage(hrq rq)
{
    char *qmark, *file, *dot, *user, *pass, *path;

    /* parse variables from form submission */
    if(j_strcmp(xmlnode_get_attrib(rq->x, "type"),"post") == 0)
        http_qs2xml(rq->body,xmlnode_get_data(rq->body));

    /* trim the :port from the host header */
    file = xmlnode_get_attrib(rq->head,"host");
    if(file != NULL && (dot = strstr(file,":")) != NULL)
        *dot = '\0';

    /* clean up the path, remove any query string */
    path = xmlnode_get_attrib(rq->x,"path");
    qmark = strstr(path,"?");
    if(qmark != NULL)
    {
        *qmark = '\0';
        qmark++;
        http_qs2xml(rq->body,qmark);
    }
    http_strdecode(path);

    /* decode any basic auth */
    if(j_strncmp(xmlnode_get_attrib(rq->head,"authorization"),"Basic ",6) == 0)
    {
        user = pass = NULL;
        user = strstr(xmlnode_get_attrib(rq->head,"authorization")," ");
        user++;
        str_b64decode(user);
        if(user != NULL)
            pass = strchr(user,':');
        if(pass != NULL)
        {
            *pass = '\0';
            pass++;
        }
        log_debug(ZONE,"Decoded %s / %s",user,pass);
        xmlnode_put_attrib(rq->head,"username",user);
        xmlnode_put_attrib(rq->head,"password",pass);
    }

    xmlnode_put_attrib(rq->x,"ip",mio_ip(rq->m));
}

void http_in_route(hrq rq, char *type)
{
    xmlnode r;

    r = xmlnode_wrap(rq->x,"route");
    xmlnode_put_attrib(r,"to",xmlnode_get_attrib(rq->x,"to"));
    xmlnode_put_attrib(r,"from",xmlnode_get_attrib(rq->x,"from"));
    xmlnode_put_attrib(r,"type",type);
    deliver(dpacket_new(xmlnode_dup(r)),NULL);
}

/* socket disconnects */
void http_in_cleanup(void *arg)
{
    hrq rq = (hrq)arg;

    if(rq->x == NULL) return;

    /* inform the component */
    http_in_route(rq,"error");
}

void http_in_reply(hrq rq, char *type, xmlnode x)
{
    char len[10], *body, *status;
    spool s;
    xmlnode att, xbody, head;

    status = xmlnode_get_attrib(x,"status");
    head = xmlnode_get_tag(x,"head");
    xbody = xmlnode_get_tag(x,"body");

    /* ugh, error!  be nice */
    if(j_strcmp(type,"error") == 0)
    {
        log_notice(rq->h->i->id,"dropping http request to %s from %s",xmlnode_get_attrib(rq->x,"to"),mio_ip(rq->m));
        status = pstrdup(rq->p,"500 Internal Error");
        head = xmlnode_new_tag_pool(rq->p,"head");
        xbody = xmlnode_new_tag_pool(rq->p,"body");
        xmlnode_insert_cdata(xbody,"<h1>Internal Error: Unable to parse request</h1>",-1);
    }

    if(j_strcmp(type,"close") == 0)
    {
        rq->x = NULL;
        mio_close(rq->m);
        return;
    }

    s = spool_new(rq->p);
    if(xmlnode_get_type(xmlnode_get_firstchild(xbody)) == NTYPE_TAG)
        body = xmlnode2str(xmlnode_get_firstchild(xbody));
    else
        body = xmlnode_get_data(xbody);

    /* status response */
    if(status != NULL)
        spooler(s, "HTTP/1.0 ", status, "\r\n", s);

    /* default and dynamic headers */
    if(head != NULL)
    {
        spool_add(s, "Server: jabber/http-" VERSION " \r\n");
        spool_add(s, "Expires: Fri, 10 Oct 1997 10:10:10 GMT\r\nPragma: no-cache\r\nCache-control: private\r\nConnection: close\r\n");
        for(att = xmlnode_get_firstattrib(head); att != NULL; att = xmlnode_get_nextsibling(att))
            spooler(s, xmlnode_get_name(att), ": ", xmlnode_get_data(att), "\r\n", s);

        if(j_strcmp(type,"partial") != 0)
        {
            sprintf(len,"%d",j_strlen(body));
            spooler(s,"Content-length: ",len,"\r\n",s);
        }
        spool_add(s, "\r\n");
    }
    spool_add(s,body);

    body = spool_print(s);
    log_debug(ZONE,"HTTP response [%s]",body);
    mio_write(rq->m, NULL, body, -1);

    /*  type="partial" writes and doesn't close, waiting for more <body/> sections */
    if(j_strcmp(type,"partial") == 0)
    {
        if(xmlnode_get_attrib(x,"keepalive") != NULL)
            rq->keepalive = pstrdup(rq->p,xmlnode_get_attrib(x,"keepalive"));
        return;
    }

    rq->x = NULL;
    mio_close(rq->m);
}

void http_in_process(hrq rq)
{
    char *to;
    jid from;

    if(rq->phase == PHASE_ERROR)
    {
        log_notice(rq->h->i->id,"unable to parse request from %s",mio_ip(rq->m));
        http_in_reply(rq,"error",NULL);
        return;
    }

    http_massage(rq);
    pool_cleanup(rq->p, http_in_cleanup, (void *)rq); /* if we get disconnected abruptly */

    if((to = xhash_get(rq->l->map, xmlnode_get_attrib(rq->head,"host"))) == NULL)
        if((to = rq->l->def) == NULL)
        {
            log_notice(rq->h->i->id,"no component found to handle http request to '%s', dropping",xmlnode_get_attrib(rq->head,"host"));
            rq->x = NULL;
            http_in_reply(rq,"error",NULL);
            return;
        }

    from = jid_new(rq->p, rq->h->i->id);
    jid_set(from, rq->id, JID_USER);
    xmlnode_put_attrib(rq->x, "to", to);
    xmlnode_put_attrib(rq->x, "from", jid_full(from));

    log_debug(ZONE,"got a request we're forwarding: %s",xmlnode2str(rq->x));
    http_in_route(rq,"request");
}


/* we handle incoming connections and feed right to read */
void http_in_connect(mio m, int flag, void *arg, char *buffer, int bufsz)
{
    hl l = (hl)arg;
    hrq rq;

    if(flag != MIO_NEW) return;

    log_debug(ZONE,"new incoming http connection established: %s",mio_ip(m));
    rq = http_read(l->h, m, http_in_process);
    rq->l = l; /* save our listen stuff for l8r */
}

/* handle incoming results that go to a connected request */
void http_in(hpi h, dpacket p)
{
    hrq rq;

    log_debug(ZONE,"handling http response: %s",xmlnode2str(p->x));

    rq = xhash_get(h->readers, p->id->user);
    if(rq == NULL || rq->phase != PHASE_DONE || rq->f != http_in_process)
    { /* no request or a request that we're not watching for */
        deliver_fail(p, "No HTTP Request Found");
        return;
    }

    http_in_reply(rq, xmlnode_get_attrib(p->x,"type"), xmlnode_get_firstchild(p->x));
    xmlnode_free(p->x);
}

