/* --------------------------------------------------------------------------
 *
 * License
 *
 * The contents of this file are subject to the Jabber Open Source License
 * Version 1.0 (the "License").  You may not copy or use this file, in either
 * source code or executable form, except in compliance with the License.  You
 * may obtain a copy of the License at http://www.jabber.com/license/ or at
 * http://www.opensource.org/.  
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied.  See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * Copyrights
 * 
 * Portions created by or assigned to Jabber.com, Inc. are 
 * Copyright (c) 1999-2000 Jabber.com, Inc.  All Rights Reserved.  Contact
 * information for Jabber.com, Inc. is available at http://www.jabber.com/.
 *
 * Portions Copyright (c) 1998-1999 Jeremie Miller.
 * 
 * Acknowledgements
 * 
 * Special thanks to the Jabber Open Source Contributors for their
 * suggestions and support of Jabber.
 * 
 * --------------------------------------------------------------------------*/
 
#include "http.h"

typedef struct ho_struct
{
    hpi h;
    xmlnode x;
    char *ip;
    char *ipn;
    hrq rq;
} *ho, _ho;

void http_out_route(ho o, char *type)
{
    xmlnode x;

    /* inform the requestor */
    x = xmlnode_wrap(o->x,"route");
    xmlnode_put_attrib(x,"type",type);
    xmlnode_put_attrib(x,"to",xmlnode_get_attrib(o->x,"from"));
    xmlnode_put_attrib(x,"from",o->h->i->id);
    deliver(dpacket_new(x),NULL);
}

/* socket disconnects, now we return whatever they had */
void http_out_cleanup(void *arg)
{
    ho o = (ho)arg;

    if(o->x == NULL) return;

    if(o->rq == NULL || o->rq->phase != PHASE_DONE)
    {
        http_out_route(o, "error");
        return;
    }

    /* replace request with result, copying sender and internal route baggage */
    xmlnode_put_attrib(o->rq->x, "from", xmlnode_get_attrib(o->x,"from"));
    xmlnode_insert_tag_node(o->rq->x, xmlnode_get_tag(o->x,"route"));
    o->x = xmlnode_dup_pool(xmlnode_pool(o->x),o->rq->x);
    http_out_route(o, "result");
}

void http_out_process(hrq rq)
{
    log_debug(ZONE,"done reading request to %s, got: %s",mio_ip(rq->m),xmlnode2str(rq->x));
    mio_close(rq->m);
}


void http_out_request(ho o, mio m)
{
    char len[10], *body, *type, *path;
    spool s;
    xmlnode att, xbody, head;
    pool p = xmlnode_pool(o->x);

    type = xmlnode_get_attrib(o->x,"type");
    path = xmlnode_get_attrib(o->x,"path");

    if(type == NULL || path == NULL)
        return;

    while(*type != '\0')
        *type++ = toupper(*type);
    type = xmlnode_get_attrib(o->x,"type");
    head = xmlnode_get_tag(o->x,"head");
    xbody = xmlnode_get_tag(o->x,"body");

    s = spool_new(p);
    if(xmlnode_get_type(xmlnode_get_firstchild(xbody)) == NTYPE_TAG)
        body = xmlnode2str(xmlnode_get_firstchild(xbody));
    else
        body = xmlnode_get_data(xbody);

    spooler(s, type, " ", path, " HTTP/1.0\r\n", s);

    if(xmlnode_get_attrib(head,"User-Agent") == NULL)
        spool_add(s, "User-Agent: jabber/http-" VERSION " \r\n");
    spool_add(s, "Expires: Fri, 10 Oct 1997 10:10:10 GMT\r\nPragma: no-cache\r\nCache-control: private\r\nConnection: close\r\n");
    for(att = xmlnode_get_firstattrib(head); att != NULL; att = xmlnode_get_nextsibling(att))
        spooler(s, xmlnode_get_name(att), ": ", xmlnode_get_data(att), "\r\n", s);

    if(body != NULL)
    {
        sprintf(len,"%d",j_strlen(body));
        spooler(s,"Content-length: ",len,"\r\n",s);
        spool_add(s, "\r\n");
        spool_add(s,body);
    }else{
        spool_add(s, "\r\n");
    }

    body = spool_print(s);
    log_debug(ZONE,"HTTP request [%s]",body);
    mio_write(m, NULL, body, -1);
}

/* we handle outgoing connections */
void http_out_connect(mio m, int flag, void *arg, char *buffer, int bufsz)
{
    ho o = (ho)arg;

    if(flag == MIO_CLOSED && o->rq == NULL)
    {
        if(o->ipn != NULL)
        {
            /* XXX try next one */
            return;
        }
        log_debug(ZONE,"unable to connect to %s, giving up",o->ip);
        http_out_route(o,"error");
        return;
    }
    if(flag != MIO_NEW) return;

    log_debug(ZONE,"new outgoing http connection established: %s",mio_ip(m));
    http_out_request(o, m);
    o->rq = http_read(o->h, m, http_out_process);
    pool_cleanup(mio_pool(m), http_out_cleanup, (void *)o); /* sends the result back to the requestor */
}

/* handle outgoing requests, connect to a host */
void http_out(hpi h, dpacket p)
{
    ho o;
    int port = 80;

    log_debug(ZONE,"handling http request: %s",xmlnode2str(p->x));

    if(j_strcmp(xmlnode_get_attrib(p->x,"type"),"error") == 0)
    { /* errors? */
        xmlnode_free(p->x);
        return;
    }

    o = pmalloco(xmlnode_pool(p->x), sizeof(_ho));
    o->x = xmlnode_get_firstchild(p->x);
    o->ip = xmlnode_get_attrib(p->x,"ip");
    o->h = h;
    /* XXX split up ip:port;ip:port */
    mio_connect(o->ip, j_atoi(xmlnode_get_attrib(o->x,"port"),port), http_out_connect, (void *)o, 30, NULL, NULL);
    
}




