/* --------------------------------------------------------------------------
 *
 * License
 *
 * The contents of this file are subject to the Jabber Open Source License
 * Version 1.0 (the "License").  You may not copy or use this file, in either
 * source code or executable form, except in compliance with the License.  You
 * may obtain a copy of the License at http://www.jabber.com/license/ or at
 * http://www.opensource.org/.  
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied.  See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * Copyrights
 * 
 * Portions created by or assigned to Jabber.com, Inc. are 
 * Copyright (c) 1999-2000 Jabber.com, Inc.  All Rights Reserved.  Contact
 * information for Jabber.com, Inc. is available at http://www.jabber.com/.
 *
 * Portions Copyright (c) 1998-1999 Jeremie Miller.
 * 
 * Acknowledgements
 * 
 * Special thanks to the Jabber Open Source Contributors for their
 * suggestions and support of Jabber.
 * 
 * --------------------------------------------------------------------------*/
 
#include "http.h"


void http_read_error(hrq rq)
{
    log_debug(ZONE,"read error, old phase %d xml %s buffer %s",rq->phase, xmlnode2str(rq->x), rq->buffer);
    rq->phase = PHASE_ERROR;
    if(rq->f != NULL)
        (rq->f)(rq);
}

/* parse the start line, either:
    HTTP/1.1 200 OK
or
    GET / HTTP/1.0
*/
void _scan_start(hrq rq)
{
    char *eol, *mid, *end, *str;

    /* get end of line or fail */
    eol = strstr(rq->buffer,"\r\n");
    if(eol == NULL) return;
    *eol = '\0';
    eol += 2;

    /* get the middle */
    mid = strstr(rq->buffer," ");
    if(mid == NULL) return http_read_error(rq);
    *mid = '\0';
    ++mid;

    rq->x = xmlnode_new_tag_pool(rq->p,"http");
    rq->head = xmlnode_insert_tag(rq->x,"head");
    rq->body = xmlnode_insert_tag(rq->x,"body");
    if(strncmp(rq->buffer,"HTTP/1",6) == 0)
    {
        xmlnode_put_attrib(rq->x,"type","result");
        xmlnode_put_attrib(rq->x,"status",mid);
    }else{
    
        /* split off the protocol */
        end = strstr(mid," HTTP/1");
        if(end == NULL) return http_read_error(rq);
        *end = '\0';
        ++end;

        /* lowercase the request type */
        for(str = rq->buffer; *str != '\0'; str++)
        *str = tolower(*str);
        xmlnode_put_attrib(rq->x,"type",rq->buffer);
        xmlnode_put_attrib(rq->x,"path",mid);
    }

    /* next step */
    rq->buffer = eol;
    rq->phase = PHASE_HEAD;
}

void _scan_head(hrq rq)
{
    char *eol, *header, *str;

    for(eol = strstr(rq->buffer,"\r\n"); eol != NULL; eol = strstr(rq->buffer,"\r\n"))
    {
        if(eol == rq->buffer)
        {
            rq->buffer += 2; /* move past eol */
            rq->clen = j_atoi(xmlnode_get_attrib(rq->head,"content-length"),0);
            if(rq->clen > 0)
                rq->phase = PHASE_BODY;
            else
                rq->phase = PHASE_DONE;
            return;
        }

        /* chop off */
        *eol = '\0';
        eol += 2;

        /* segment header/value */
        header = strstr(rq->buffer,": ");
        if(header == NULL) return http_read_error(rq);
        *header = '\0';
        header += 2;

        /* lowercase the header names (IE SUCKS!) */
        for(str = rq->buffer; *str != '\0'; str++)
            *str = tolower(*str);
        xmlnode_put_attrib(rq->head,rq->buffer,header);
        rq->buffer = eol;
    }
}

void _scan_body(hrq rq)
{
    xmlnode x;

    if(strlen(rq->buffer) < rq->clen) return;

    if(j_strcmp(xmlnode_get_attrib(rq->head,"content-type"),"text/xml") == 0)
    {
        if((x = xmlnode_str(rq->buffer, rq->clen)) == NULL) return http_read_error(rq);
        xmlnode_insert_tag_node(rq->body,x);
        xmlnode_free(x);
    }else{
        xmlnode_insert_cdata(rq->body,rq->buffer,rq->clen);
    }
    rq->phase = PHASE_DONE;
}

/* read in the http request */
void http_read_eat(hrq rq, void *buffer, int bufsz)
{
    char *cur;

    if(bufsz <= 0) return;

    cur = pmalloc(rq->p,strlen(rq->buffer) + bufsz + 1);
    *cur = '\0';
    strcat(cur,rq->buffer);
    strcat(cur,buffer);
    rq->buffer = cur;

    log_debug(ZONE,"reading data [%s]",rq->buffer);

    if(rq->phase == PHASE_START) _scan_start(rq);
    if(rq->phase == PHASE_HEAD) _scan_head(rq);
    if(rq->phase == PHASE_BODY) _scan_body(rq);

    /* lookup and deliver the request */
    if(rq->phase == PHASE_DONE)
    {
        log_debug(ZONE,"got request: %s",xmlnode2str(rq->x));
        (rq->f)(rq);
    }
}

void http_read_cleanup(void *arg)
{
    hrq rq = (hrq)arg;

    xhash_zap(rq->h->readers, rq->id);
}

/* we handle incoming data on the socket */
void http_read_io(mio m, int flag, void *arg, char *buffer, int bufsz)
{
    hrq rq = (hrq)arg;

    if(rq->phase >= PHASE_DONE) return;

    if(flag == MIO_BUFFER)
        http_read_eat(rq, buffer, bufsz);
}

hrq http_read(hpi h, mio m, httpreader f)
{
    hrq rq;

    log_debug(ZONE,"new http reader created on: %s",mio_ip(m));
    rq = pmalloco(mio_pool(m), sizeof(_hrq));
    rq->h = h;
    snprintf(rq->id,10,"%X",m); /* unique id */
    xhash_put(rq->h->readers,rq->id,(void *)rq);
    rq->m = m;
    rq->p = mio_pool(m);
    pool_cleanup(rq->p,http_read_cleanup,(void *)rq);
    rq->stamp = time(NULL);
    rq->buffer = pstrdup(rq->p,"");
    rq->f = f;
    mio_reset(m, http_read_io, (void*)rq);
    return rq;
}


