/* --------------------------------------------------------------------------
 *
 * License
 *
 * The contents of this file are subject to the Jabber Open Source License
 * Version 1.0 (the "License").  You may not copy or use this file, in either
 * source code or executable form, except in compliance with the License.  You
 * may obtain a copy of the License at http://www.jabber.com/license/ or at
 * http://www.opensource.org/.  
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied.  See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * Copyrights
 * 
 * Portions created by or assigned to Jabber.com, Inc. are 
 * Copyright (c) 1999-2000 Jabber.com, Inc.  All Rights Reserved.  Contact
 * information for Jabber.com, Inc. is available at http://www.jabber.com/.
 *
 * Portions Copyright (c) 1998-1999 Jeremie Miller.
 * 
 * Acknowledgements
 * 
 * Special thanks to the Jabber Open Source Contributors for their
 * suggestions and support of Jabber.
 * 
 * --------------------------------------------------------------------------*/
 
#include "jabberd.h"

/*

This is a simple test for the http component, map any http traffic to it and it'll respond w/ a simple page

  <service id="httptest">
    <load><httptest>./http/httptest.so</httptest></load>
  </service>

*/

/* the callback to handle incoming packets */
result httptest_packets(instance i, dpacket p, void *arg)
{
    xmlnode x2, x = xmlnode_get_firstchild(p->x);
    char *type = xmlnode_get_attrib(p->x,"type");
    char *rqbody = NULL;
    xmlnode body;

    log_debug(ZONE,"incoming http test packet %s",xmlnode2str(p->x));

    if(j_strcmp(type,"result") == 0)
    {
        log_debug(ZONE,"result from our request");
        rqbody = spools(p->p,"<H1>get test returned:</H1><HR>",xmlnode_get_tag_data(x,"body"),p->p);
        p->x = xmlnode_get_tag(x,"route");
        x = xmlnode_get_firstchild(p->x);
    }else if(j_strcmp(type,"request") != 0){
        log_debug(ZONE,"error!");
        xmlnode_free(p->x);
        return r_DONE;
    }

    if(rqbody == NULL && j_strcmp(xmlnode_get_attrib(x,"path"),"/get") == 0)
    {
        /* dummy request */
        x2 = xmlnode_new_tag("http");
        xmlnode_put_attrib(x2,"type","get");
        x = xmlnode_get_tag(x,"body");

        /* get vars from request for other host than us */
        if(xmlnode_get_attrib(x,"path") != NULL)
            xmlnode_put_attrib(x2,"path",xmlnode_get_attrib(x,"path"));
        else
            xmlnode_put_attrib(x2,"path","/");
        if(xmlnode_get_attrib(x,"to") != NULL)
            xmlnode_put_attrib(x2,"to",xmlnode_get_attrib(x,"to"));
        else
            xmlnode_put_attrib(x2,"to","localhost");
        if(xmlnode_get_attrib(x,"port") != NULL)
            xmlnode_put_attrib(x2,"port",xmlnode_get_attrib(x,"port"));
        else
            xmlnode_put_attrib(x2,"port","5288");

        /* sneak the orig request in on the payload */
        xmlnode_insert_tag_node(x2,p->x);

        x = xmlnode_wrap(x2,"route");
        xmlnode_put_attrib(x,"to","http");
        xmlnode_put_attrib(x,"from","httptest");
        xmlnode_put_attrib(x,"type","request");
        deliver(dpacket_new(x),NULL);
        return r_DONE;
    }

    jutil_tofrom(p->x);
    xmlnode_put_attrib(p->x, "type", "result");
    xmlnode_put_attrib(x, "status", "200 Ok");
    xmlnode_hide(xmlnode_get_tag(x,"head"));
    xmlnode_hide(xmlnode_get_tag(x,"body"));
    body = xmlnode_insert_tag(x, "body");

    if(j_strcmp(xmlnode_get_attrib(x,"path"),"/xml") == 0)
    {
        xmlnode_put_attrib(xmlnode_insert_tag(x, "head"), "Content-type", "text/xml");
        body = xmlnode_insert_tag(body,"xml");
        xmlnode_put_attrib(xmlnode_insert_tag(body,"foo"),"bar","foobar");
        xmlnode_insert_cdata(xmlnode_insert_tag(body,"bar"),"foobar",-1);
    }else if(rqbody != NULL){
        xmlnode_put_attrib(xmlnode_insert_tag(x, "head"), "Content-type", "text/html");
        xmlnode_insert_cdata(body, rqbody, -1);
    }else{
        xmlnode_put_attrib(xmlnode_insert_tag(x, "head"), "Content-type", "text/html");
        xmlnode_insert_cdata(body, "
<H1>httptest component</H1>
<a href='/xml'>xml test</a><br>
<a href='/get'>get test (internally fetches this page via http)</a><br>
", -1);
    }

    deliver(dpacket_new(p->x),i);
    return r_DONE;
}


/********** startup/main function ************/

void httptest(instance i, xmlnode x)
{
    log_debug(ZONE,"httptest loading");

    register_phandler(i, o_DELIVER, httptest_packets, NULL);
}

